using System;
using System.IO;
using System.Diagnostics;
using System.Configuration;

namespace GridWorld
{
    class EnvironmentException : Exception
    {
        public EnvironmentException() : base() { }
        public EnvironmentException(string message) : base(message) { }
        public EnvironmentException(string message, Exception innerException) : base(message, innerException) { }
    }

    internal struct Cell
    {
        public int i;
        public int j;
        public int reward;
        public bool isAbsorbing;
    }

    class Environment : IEnvironment
    {
        private Cell[,] field;
        private int numLines;
        private int numColumns;

        // for experimental part
        private static Random rnd = new Random();
        private int[,] rewards;
        private const int stepCounter = 10;
        private int[,] visitings;
        private bool[,] used;

        #region Initialization

        public Environment()
        {
            // initialize field array and its dimensions
            CreateField("test.txt");
        }

        private void CreateField(string fileName)
        {
            StreamReader sr = null;
            try
            {
                sr = new StreamReader(fileName);

                string[] line = GetNextLine(sr, 2);
                numLines = int.Parse(line[0]);
                numColumns = int.Parse(line[1]);
                if (numLines <= 0 || numColumns <= 0)
                    throw new EnvironmentException("Incorrect input data - field dimensions must be positive integers");

                field = new Cell[numLines, numColumns];
                FillArray(sr);
            }
            catch (IOException)
            {
                Console.WriteLine("Can't load test file.");
                System.Environment.Exit(1);
            }
            finally { if (sr != null) sr.Close(); }
        }

        private static string[] GetNextLine(StreamReader sr, int expectedLength)
        {
            string[] line = sr.ReadLine().Split(new char[] { ' ' }, StringSplitOptions.RemoveEmptyEntries);
            Debug.Assert(line.Length == expectedLength);

            return line;
        }

        private void FillArray(StreamReader sr)
        {
            for (int i = 0; i < numLines; i++)
            {
                string[] line = GetNextLine(sr, numColumns);

                for (int j = 0; j < numColumns; j++)
                {
                    if (line[j].Contains("("))
                    {
                        string[] tmp = line[j].Split(
                            new char[] { '(', ',', ')', ' ' },
                            StringSplitOptions.RemoveEmptyEntries);
                        Debug.Assert(tmp.Length == 3);

                        field[i, j].isAbsorbing = true;
                        field[i, j].reward = int.Parse(tmp[0]);
                        field[i, j].i = int.Parse(tmp[1]);
                        field[i, j].j = int.Parse(tmp[2]);
                    }
                    else
                    {
                        field[i, j].reward = int.Parse(line[j]);
                    }
                }
            }
        }

        #endregion // Initialization

        public IAgent CreateAgent()
        {
            return new QLearningAgent(this, numLines, numColumns);
        }

        #region Training Phase

        private int CheckAbsorbingState(ref int i, ref int j)
        {
            int absorbingReward = 0;
            Cell cell = field[i - 1, j - 1];
            if (cell.isAbsorbing)
            {
                absorbingReward = cell.reward;
                i = cell.i;
                j = cell.j;
            }

            return absorbingReward;
        }

        /// <summary>
        /// Performs agent's action which changes current state under environment rules.
        /// </summary>
        /// <param name="state">Current state</param>
        /// <param name="action">Chosen action</param>
        /// <returns>Reward for performed action</returns>
        public int PerformAction(ref int state, int action)
        {
            int i = state / numColumns + 1;
            int j = state % numColumns + 1;
            int absorbingReward = 0;
            const int wallPenalty = -100;

            switch (action)
            {
                case 0: // up
                    if (i == 1)
                        return wallPenalty;
                    i = (i - 1);
                    absorbingReward = CheckAbsorbingState(ref i, ref j);
                    break;
                case 1: // right
                    if (j == numColumns)
                        return wallPenalty;
                    j = (j + 1);
                    absorbingReward = CheckAbsorbingState(ref i, ref j);
                    break;
                case 2: // down
                    if (i == numLines)
                        return wallPenalty;
                    i = (i + 1);
                    absorbingReward = CheckAbsorbingState(ref i, ref j);
                    break;
                case 3: // left
                    if (j == 1)
                        return wallPenalty;
                    j = (j - 1);
                    absorbingReward = CheckAbsorbingState(ref i, ref j);
                    break;
                default:
                    throw new EnvironmentException("Incorrect action.");
            }

            state = (i - 1) * numColumns + j - 1;
            return field[i - 1, j - 1].reward + absorbingReward;
        }

        #endregion // Training Phase

        #region Experimental Phase

        /// <summary>
        /// Analyzes behavior of agent-generated stochastic matrix.
        /// </summary>
        /// <param name="matrix"></param>
        public void PerformExperiment(double[,] matrix)
        {
            visitings = new int[numLines, numColumns];

            PrintAvarageRewards(matrix);

            PrintCompressionResults(matrix);
        }

        #region Methods for calculation avarage rewards

        private void PrintAvarageRewards(double[,] matrix)
        {
            const int numOfTests = 50;
            int numOfMoves = Convert.ToInt32(ConfigurationManager.AppSettings["NumOfMoves"]);

            for (int numOfAgents = 1; numOfAgents <= 40; numOfAgents++)
            {
                int totalReward = 0;
                for (int k = 0; k < numOfTests; k++)
                {
                    int[] states = new int[numOfAgents];
                    InitializeRewards();

                    for (int i = 0; i < numOfMoves; i++)
                    {
                        for (int j = 0; j < numOfAgents; j++)
                        {
                            int action = GetNextAction(matrix, states[j]);
                            totalReward += PerformMove2(ref states[j], action);
                        }

                        UpdateRewards();
                    }
                }

                Console.WriteLine("Average reward for {0} agents: {1}", numOfAgents, totalReward / numOfTests);
            }
            Console.WriteLine();
        }

        private int CheckAbsorbingState2(ref int i, ref int j)
        {
            int absorbingReward = 0;
            Cell cell = field[i - 1, j - 1];
            if (cell.isAbsorbing)
            {
                if (rewards[i - 1, j - 1] >= stepCounter)
                {
                    absorbingReward = cell.reward;
                    rewards[i - 1, j - 1] = 0;
                }
                i = cell.i;
                j = cell.j;
                visitings[i - 1, j - 1]++;
            }

            return absorbingReward;
        }

        private int PerformMove2(ref int state, int action)
        {
            int i = state / numColumns + 1;
            int j = state % numColumns + 1;
            int absorbingReward;
            const int wallPenalty = -100;

            switch (action)
            {
                case 0: // up
                    if (i == 1)
                        return wallPenalty;
                    i = (i - 1);
                    visitings[i - 1, j - 1]++;
                    absorbingReward = CheckAbsorbingState2(ref i, ref j);
                    break;
                case 1: // right
                    if (j == numColumns)
                        return wallPenalty;
                    j = (j + 1);
                    visitings[i - 1, j - 1]++;
                    absorbingReward = CheckAbsorbingState2(ref i, ref j);
                    break;
                case 2: // down
                    if (i == numLines)
                        return wallPenalty;
                    i = (i + 1);
                    visitings[i - 1, j - 1]++;
                    absorbingReward = CheckAbsorbingState2(ref i, ref j);
                    break;
                case 3: // left
                    if (j == 1)
                        return wallPenalty;
                    j = (j - 1);
                    visitings[i - 1, j - 1]++;
                    absorbingReward = CheckAbsorbingState2(ref i, ref j);
                    break;
                default:
                    throw new EnvironmentException("Incorrect action.");
            }

            state = (i - 1) * numColumns + j - 1;

            int reward = 0;
            if (rewards[i - 1, j - 1] >= stepCounter && field[i - 1, j - 1].reward > 0)
            {
                reward = field[i - 1, j - 1].reward;
                rewards[i - 1, j - 1] = 0;
            }
            return reward + absorbingReward;
        }

        private void InitializeRewards()
        {
            rewards = new int[numLines, numColumns];
            for (int i = 0; i < numLines; i++)
            {
                for (int j = 0; j < numColumns; j++)
                {
                    if (field[i, j].reward > 0) rewards[i, j] = stepCounter;
                }
            }
        }

        private void UpdateRewards()
        {
            for (int i = 0; i < numLines; i++)
            {
                for (int j = 0; j < numColumns; j++)
                {
                    if (field[i, j].reward > 0 && rewards[i, j] < stepCounter)
                        rewards[i, j]++;
                }
            }
        }

        private int GetNextAction(double[,] matrix, int state)
        {
            int action = 0;
            double random = rnd.NextDouble();
            double sum = 0.0;
            const int numOfActions = 4;

            while (action < numOfActions - 1)
            {
                sum += matrix[state, action];
                if (random < sum)
                    break;

                action++;
            }

            return action;
        }

        #endregion // Methods for calculation avarage rewards

        #region Methods for research compression

        private void PrintCompressionResults(double[,] matrix)
        {
            PrintMatrix(visitings);
            InitializeUsedMatrix();
            int numOfMoves = Convert.ToInt32(ConfigurationManager.AppSettings["NumOfMoves"]);

            for (int t = 0; t < 15; t++)
            {
                KillNextMinimumState();
                const int numOfTests = 50;

                for (int numOfAgents = 1; numOfAgents <= 40; numOfAgents++)
                {
                    int totalReward = 0;
                    for (int k = 0; k < numOfTests; k++)
                    {
                        int[] states = new int[numOfAgents];
                        InitializeRewards();

                        for (int i = 0; i < numOfMoves; i++)
                        {
                            for (int j = 0; j < numOfAgents; j++)
                            {
                                int action = GetNextAction(matrix, states[j]);
                                while (!ActionAvailable(states[j], action))
                                    action = GetNextAction(matrix, states[j]);
                                totalReward += PerformMove3(ref states[j], action);
                            }

                            UpdateRewards();
                        }
                    }

                    Console.WriteLine("Average reward for {0} agents: {1}", numOfAgents, totalReward / numOfTests);
                }
                Console.WriteLine();
            }
        }

        private bool ActionAvailable(int state, int action)
        {
            int i = state / numColumns + 1;
            int j = state % numColumns + 1;

            switch (action)
            {
                case 0: // up
                    if (i == 1)
                        throw new EnvironmentException("Incorrect action.");
                    i = (i - 1);
                    break;
                case 1: // right
                    if (j == numColumns)
                        throw new EnvironmentException("Incorrect action.");
                    j = (j + 1);
                    break;
                case 2: // down
                    if (i == numLines)
                        throw new EnvironmentException("Incorrect action.");
                    i = (i + 1);
                    break;
                case 3: // left
                    if (j == 1)
                        throw new EnvironmentException("Incorrect action.");
                    j = (j - 1);
                    break;
                default:
                    throw new EnvironmentException("Incorrect action.");
            }

            bool result = used[i - 1, j - 1];
            Cell cell = field[i - 1, j - 1];
            if (cell.isAbsorbing)
            {
                if (!used[cell.i - 1, cell.j - 1]) result = false;
            }

            return result;
        }

        private void InitializeUsedMatrix()
        {
            used = new bool[numLines, numColumns];

            for (int i = 0; i < numLines; i++)
            {
                for (int j = 0; j < numColumns; j++)
                {
                    used[i, j] = true;
                }
            }
        }

        private void KillNextMinimumState()
        {
            int minI = 0, minJ = 0;
            for (int i = 0; i < numLines; i++)
            {
                for (int j = 0; j < numColumns; j++)
                {
                    if (used[i, j] && field[i, j].isAbsorbing == false
                        && visitings[minI, minJ] > visitings[i, j])
                    {
                        minI = i;
                        minJ = j;
                    }
                }
            }

            if (used[minI, minJ])
                Console.WriteLine("State [{0}, {1}] was successfully removed", minI, minJ);
            used[minI, minJ] = false;
        }

        private int CheckAbsorbingState3(ref int i, ref int j)
        {
            int absorbingReward = 0;
            Cell cell = field[i - 1, j - 1];
            if (cell.isAbsorbing)
            {
                if (rewards[i - 1, j - 1] >= stepCounter)
                {
                    absorbingReward = cell.reward;
                    rewards[i - 1, j - 1] = 0;
                }
                i = cell.i;
                j = cell.j;
            }

            return absorbingReward;
        }

        private int PerformMove3(ref int state, int action)
        {
            int i = state / numColumns + 1;
            int j = state % numColumns + 1;
            int absorbingReward;
            const int wallPenalty = -100;

            switch (action)
            {
                case 0: // up
                    if (i == 1)
                        return wallPenalty;
                    i = (i - 1);
                    absorbingReward = CheckAbsorbingState3(ref i, ref j);
                    break;
                case 1: // right
                    if (j == numColumns)
                        return wallPenalty;
                    j = (j + 1);
                    absorbingReward = CheckAbsorbingState3(ref i, ref j);
                    break;
                case 2: // down
                    if (i == numLines)
                        return wallPenalty;
                    i = (i + 1);
                    absorbingReward = CheckAbsorbingState3(ref i, ref j);
                    break;
                case 3: // left
                    if (j == 1)
                        return wallPenalty;
                    j = (j - 1);
                    absorbingReward = CheckAbsorbingState3(ref i, ref j);
                    break;
                default:
                    throw new EnvironmentException("Incorrect action.");
            }

            state = (i - 1) * numColumns + j - 1;

            int reward = 0;
            if (rewards[i - 1, j - 1] >= stepCounter && field[i - 1, j - 1].reward > 0)
            {
                reward = field[i - 1, j - 1].reward;
                rewards[i - 1, j - 1] = 0;
            }
            return reward + absorbingReward;
        }

        #endregion // Methods for research compression

        #endregion

        private void PrintMatrix(int[,] matrix)
        {
            for (int i = 0; i < numLines; i++)
            {
                for (int j = 0; j < numColumns; j++)
                {
                    Console.Write(matrix[i, j] + " ");
                }
                Console.WriteLine();
            }
            Console.WriteLine();
        }

    }
}
