using System;
using System.Configuration;
using System.Diagnostics;
using System.IO;

namespace GridWorld
{
    class QLearningAgent : IAgent
    {
        // current task
        private readonly IEnvironment env;
        // number of lines in Grid-world environment task
        private readonly int numLines;
        // number of columns in Grid-world environment task
        private readonly int numColumns;
        // q-function, represents potential rewards for all <state, action> pairs
        private double[,] q;
        // zero-based state of agent
        private int state;

        // number of moves for one training phase
        private static readonly int numOfMoves;
        // q-learning step-size parameter
        private static readonly double stepSize;
        // number of actions for each state
        private const int numOfActions = 4;

        private static Random rnd = new Random();

        #region Constructors

        static QLearningAgent()
        {
            numOfMoves = Convert.ToInt32(ConfigurationManager.AppSettings["NumOfMoves"]);
            stepSize = Convert.ToDouble(ConfigurationManager.AppSettings["StepSize"]);
        }

        public QLearningAgent(IEnvironment env, int numLines, int numColumns)
        {
            Debug.Assert(env != null);
            Debug.Assert(numLines > 0 && numColumns > 0);

            this.env = env;
            this.numLines = numLines;
            this.numColumns = numColumns;
            this.state = 0; // top-left corner of the field
            q = new double[numLines * numColumns, 4];
        }

        #endregion // Constructors

        #region Supporting Methods

        private double GetMaximumQByState(int state)
        {
            double max = q[state, 0];
            for (int i = 1; i < numOfActions; i++)
            {
                if (q[state, i] > max) max = q[state, i];
            }

            return max;
        }

        private int GetActionWithMaximumQ(int state)
        {
            // TODO: fix if we have several maximum values
            int index = 0;
            for (int i = 1; i < numOfActions; i++)
            {
                if (q[state, i] > q[state, index]) index = i;
            }

            return index;
        }

        /// <summary>
        /// Prints the values of resulting q-function for diagnostic purposes.
        /// </summary>
        public void PrintModel()
        {
            using (StreamWriter sw = new System.IO.StreamWriter("output.txt", false))
            {
                for (int i = 0; i < numLines * numColumns; i++)
                {
                    if (i != 0 && i % 5 == 0)
                        sw.WriteLine();
                    sw.WriteLine("{0:F0} {1:F0} {2:F0} {3:F0}", q[i, 0], q[i, 1], q[i, 2], q[i, 3]);
                }
                double[,] matrix = GenerateStochasticMatrix();
                for (int i = 0; i < numLines * numColumns; i++)
                {
                    if (i % 5 == 0)
                        sw.WriteLine();
                    sw.WriteLine("{0:F3} {1:F3} {2:F3} {3:F3}", matrix[i, 0], matrix[i, 1], matrix[i, 2], matrix[i, 3]);
                }
            }
        }

        #endregion // Supporting Methods

        #region IAgent Members

        /// <summary>
        /// Performs one iteration of training phase.
        /// </summary>
        public void Train()
        {
            const double explorationRate = 0.3;

            for (int i = 0; i < numOfMoves; i++)
            {
                int oldState = state;

                int action = (rnd.NextDouble() <= explorationRate) ?
                    rnd.Next(numOfActions) :
                    GetActionWithMaximumQ(oldState);
                int reward = env.PerformAction(ref state, action);
                q[oldState, action] = reward + stepSize * GetMaximumQByState(state);
            }
        }

        /// <summary>
        /// Generates resulting stochastic matrix.
        /// Should be called after training phase was completed.
        /// </summary>
        /// <returns></returns>
        public double[,] GenerateStochasticMatrix()
        {
            double[,] matrix = new double[numLines * numColumns, numOfActions];

            double sum;
            for (int i = 0; i < numLines * numColumns; i++)
            {
                sum = 0;
                for (int j = 0; j < numOfActions; j++)
                {
                    if (q[i, j] > 0) sum += q[i, j];
                }
                for (int j = 0; j < numOfActions; j++)
                {
                    matrix[i, j] = (q[i, j] > 0) ? (q[i, j] / sum) : 0;
                }
            }

            return matrix;
        }

        #endregion // IAgent Members

    }
}
